# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['paroxython']

package_data = \
{'': ['*'], 'paroxython': ['resources/*']}

install_requires = \
['docopt>=0.6.2,<0.7.0',
 'regex==2020.7.14',
 'typed-ast>=1.4.1,<2.0.0',
 'typing-extensions>=3.7.4.3,<4.0.0.0']

entry_points = \
{'console_scripts': ['paroxython = paroxython.cli:main']}

setup_kwargs = {
    'name': 'paroxython',
    'version': '0.6.1',
    'description': 'Search Python code for algorithmic features',
    'long_description': '[![Build Status](https://travis-ci.com/laowantong/paroxython.svg?branch=master)](https://travis-ci.com/laowantong/paroxython)\n[![codecov](https://img.shields.io/codecov/c/github/laowantong/paroxython/master)](https://codecov.io/gh/laowantong/paroxython)\n[![Checked with mypy](https://img.shields.io/badge/typing-mypy-brightgreen)](http://mypy-lang.org/)\n[![Codacy Badge](https://api.codacy.com/project/badge/Grade/73432ed4c5294326ba6279bbbb0fe2e6)](https://www.codacy.com/manual/laowantong/paroxython)\n[![Updates](https://pyup.io/repos/github/laowantong/paroxython/shield.svg)](https://pyup.io/repos/github/laowantong/paroxython/)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/paroxython)\n[![GitHub Release](https://img.shields.io/github/release/laowantong/paroxython.svg?style=flat)]()\n![GitHub code size in bytes](https://img.shields.io/github/languages/code-size/laowantong/paroxython)\n[![paroxython SLOC](https://img.shields.io/badge/main%20program-~1850%20SLOC-blue)](https://github.com/laowantong/paroxython/blob/master/paroxython)\n[![tests SLOC](https://img.shields.io/badge/tests-~2550%20SLOC-blue)](https://github.com/laowantong/paroxython/blob/master/tests)\n[![helpers SLOC](https://img.shields.io/badge/helpers-~950%20SLOC-blue)](https://github.com/laowantong/paroxython/blob/master/helpers)\n[![spec features](https://img.shields.io/badge/spec-171%20features-blue)](https://github.com/laowantong/paroxython/blob/master/paroxython/resources/spec.md)\n[![taxonomy mappings](https://img.shields.io/badge/taxonomy-276%20mappings-blue)](https://github.com/laowantong/paroxython/blob/master/paroxython/resources/taxonomy.tsv)\n![GitHub commit activity](https://img.shields.io/github/commit-activity/y/laowantong/paroxython.svg?style=flat)\n[![Downloads](https://pepy.tech/badge/paroxython/week)](https://pepy.tech/project/paroxython/week)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![Band-aid](https://badgen.net/badge/not%C2%A0%C2%A0%F0%9F%85%B3%F0%9F%85%B4%F0%9F%85%B0%F0%9F%85%B3/yet/F3D9C5?labelColor=F3D9C5)](https://youtu.be/QcbR1J_4ICg?t=54)\n\n<p align="center">\n  <a href="https://laowantong.github.io/paroxython/index.html">\n  <img src="docs/resources/logo.png">\n  </a>\n</p>\n\n## Introduction\n\nParoxython is a set of command line tools which **tag** and **filter** by algorithmic features your collection of Python programming exercises.\n\n### Audience\n\nYou are a teacher, in charge of an introductory programming course in an educational institution. Over the years, you have accumulated many—far too many—programs and code snippets that may be of interest to your students.\n\nOr, as a seasoned developer, you would like to share your knowledge by helping a loved one learn how to code. A cursory search for pedagogical material yields an overwhelming amount of websites and repositories stuffed with Python programs of various levels (e.g.,\n[1](https://github.com/TheAlgorithms/Python),\n[2](http://rosettacode.org/wiki/Category:Python),\n[3](https://www.programming-idioms.org/about#about-block-language-coverage),\n[4](https://github.com/codebasics/py),\n[5](https://github.com/keon/algorithms),\n[6](https://github.com/OmkarPathak/Python-Programs),\nand a lot more from [Awesome Python in Education](https://github.com/quobit/awesome-python-in-education)).\n\nIn any case, the Python source codes you have gathered are typically\n**numerous** (hundreds or even thousands),\n**reasonably sized** (anything below 100 lines of code),\nand **educational** in nature (e.g., snippets, examples, quizzes, exercise solutions, classic algorithms).\nThe programming concepts you plan to teach remain relatively **low level** (e.g. assignments, nested loops, accumulation patterns, tail recursive functions, etc.).\n\nIf all that sounds familiar, keep reading me.\n\n### Main goals\n\nParoxython aims to help you select, from your collection, the one program that best suits your needs. For instance, it will gladly answer the following questions:\n\n> - How can this concept be illustrated?\n> - What problems use the same algorithmic and data structures as this one?\n> - What homework assignment should I give my students so they can practice the content of the last lesson?\n\nMoreover, since Paroxython knows what your class knows, it can recommend the right program at the right time:\n\n> - What would make a good review exercise?\n> - Which exercises can I give on this exam?\n> - What is the current learning cost of this example?\n\nIn the long run, Paroxython may guide you and somehow make you rethink your course outline:\n\n> - What are the prerequisites for the concept of assignment?\n> - Do I have enough material to introduce subroutines before I even talk about conditionals and loops?\n> - Among the loops, which must come first: the most powerful (`while`), or the most useful (`for`)?\n> - How to logically structure this bunch of usual iterative patterns?\n> - What are the _basics_, exactly?\n\nAll issues on which the author changed his mind since he started to work on this project!\n\nIn an ideal world, Paroxython could even put an end to the deadliest religious wars, with rational, data-driven arguments:\n\n> - Father, is it a sin to exit early?\n> - Should a real byte use a mask?\n\n### How it works\n\n<p align="center">\n  <a href="https://laowantong.github.io/paroxython/developer_manual/index.html">\n  <img src="docs/resources/waterfall.png">\n  </a>\n</p>\n\nParoxython starts from a given folder of **programs**. Its contents is parsed, and all features that meet the provided **specifications** are labelled and associated with their spans (e.g., `"assignment_lhs_identifier:a": 4, 6, 18` or `"loop_with_late_exit:while": 3-7, 20-29`).\n\nThese **labels** constitute only scattered knowledge. The next step is to map them onto a **taxonomy** designed with basic hierarchical constraints in mind (e.g., the fact that the introduction of the concept of early exit must come after that of loop, which itself requires that of control flow, is expressed by the _taxon_ `"flow/loop/exit/early"`).\n\n<p align="center">\n  <a href="https://laowantong.github.io/paroxython/user_manual/index.html#taxonomy">\n  <img src="docs/resources/tree.png" alt="A taxonomy.">\n  </a>\n  <br>\n  <em>Extract of the taxonomy generated from <a href="https://github.com/TheAlgorithms/Python">The Algorithms - Python</a>.<br>Click to jump to its full dynamic version in the user manual.</em>\n</p>\n\nEverything is then persisted in a tag **database**, which can later be filtered through a **pipeline** of commands, for instance:\n\n- _include_ only the programs which feature a recursive function;\n- _exclude_ this or that program you want to set aside for the exam;\n- “_impart_” all programs studied so far, _i.e_, consider that all the notions they implement are acquired.\n\nThe result is a list of program **recommendations** ordered by increasing learning cost.\n\n### Example\n\nSuppose that the `programs` directory contains [these simple programs](https://wiki.python.org/moin/SimplePrograms).\n\nFirst, build [this tag database](https://github.com/laowantong/paroxython/blob/master/examples/simple/programs_db.json):\n\n```shell\n> paroxython collect programs\nLabelling 21 programs.\nMapping taxonomy on 21 programs.\nWriting programs_db.json.\n```\n\nThen, filter it through [this pipeline](https://github.com/laowantong/paroxython/blob/master/examples/simple/programs_pipe.py):\n\n```shell\n> paroxython recommend programs\nProcessing 5 commands on 21 programs.\n  19 programs remaining after operation 1 (impart).\n  18 programs remaining after operation 2 (exclude).\n  12 programs remaining after operation 3 (exclude).\n  10 programs remaining after operation 4 (include).\n  10 programs remaining after operation 5 (hide).\nDumped: programs_recommendations.md.\n```\n\nEt voilà, [your recommendation report](https://github.com/laowantong/paroxython/blob/master/examples/simple/programs_recommendations.md)!\n\n\n## Installation and test-drive\n\n### Command line\n\nMuch to no one\'s surprise:\n\n```\npip install paroxython\n```\n\nThe following command should print a help message and exit:\n\n```\nparoxython --help\n```\n\n### IPython magic command\n\nIf you use Jupyter notebook/lab, you\'ve also just installed a so-called magic command. Load it like this:\n\n```python\n%load_ext paroxython\n```\n\nThis should print `"paroxython 0.6.1 loaded."`. Run it on a cell of Python code:\n\n```python\n%%paroxython                          # Lines\ndef fibonacci(n):                     # 2\n    result = []                       # 3\n    (a, b) = (0, 1)                   # 4\n    while a < n:                      # 5\n        result.append(a)              # 6\n        (a, b) = (b, a + b)           # 7\n    return result                     # 8\n```\n\n| Taxon | Lines |\n|:--|:--|\n| `call/subroutine/method/sequence/list/append` | 6 |\n| `condition/inequality` | 5 |\n| `def/subroutine/function/impure` | 2-8 |\n| `flow/loop/exit/late` | 5-7 |\n| `flow/loop/while` | 5-7 |\n| `meta/count/program/sloc/8` | 2-8 |\n| `meta/count/subroutine/sloc/7` | 2-8 |\n| `meta/count/variety/***` | 2-8 |\n| `meta/program` | 2-8 |\n| `operator/arithmetic/addition` | 7 |\n| `style/procedural` | 2-8 |\n| `type/number/integer/literal` | 4 |\n| `type/number/integer/literal/zero` | 4 |\n| `type/sequence/list` | 6 |\n| `type/sequence/list/literal/empty` | 3 |\n| `type/sequence/tuple/literal` | 4, 4, 7, 7 |\n| `var/assignment/explicit/parallel` | 4 |\n| `var/assignment/explicit/parallel/slide` | 7 |\n| `var/assignment/explicit/single` | 3 |\n| `var/assignment/implicit/parameter` | 2 |\n| `var/scope/local` | 2-8, 2-8, 2-8, 2-8 |\n\nAs you can see, in this program, Paroxython identifies among others:\n\n- the use of the [procedural paradigm](https://en.wikipedia.org/wiki/Procedural_programming) (`style/procedural`);\n- an im[pure function](https://en.wikipedia.org/wiki/Pure_function) (`def/subroutine/function/impure`);\n- a `while` loop (`flow/loop/while`) with a late exit (`flow/loop/exit/late`);\n- a little bit of voodoo on lists (`type/sequence/list/literal/empty` and `call/subroutine/method/sequence/list/append`);\n- a simple [tuple assignment](https://en.wikibooks.org/wiki/Python_Programming/Tuples#Packing_and_Unpacking) (`var/assignment/explicit/parallel`). Note that we distinguish between explicit (with `=`) and implicit (parameters and iteration variables) assignments;\n- a “sliding” tuple assignment (`var/assignment/explicit/parallel/slide`). If the denomination is unique to us, the pattern itself occurs in a number of programs: implementations of [C-finite sequences](https://en.wikipedia.org/wiki/Constant-recursive_sequence) with C greater than 1, [Greatest Common Divisor](https://en.wikipedia.org/wiki/Greatest_common_divisor), [Quicksort](https://en.wikipedia.org/wiki/Quicksort), etc.\n- four local variables (`var/scope/local`);\n- an estimation of the variety of concepts involved (`meta/count/variety/***`), depending on the number of lines, features and distinct features.\n\nThe magic command `%%paroxython` (corresponding to the subcommand [`tag`](https://laowantong.github.io/paroxython/cli_tag.html)) only scratches the surface of the system. As shown before, to estimate the learning cost of the features and get actionable recommendations, you will need first to construct the tag database with [`collect`](https://laowantong.github.io/paroxython/cli_collect.html), and then call [`recommend`](https://laowantong.github.io/paroxython/cli_recommend.html) on a pipeline of yours.\n\n# Read them\n\nAlthough this is still a work-in-progress, Paroxython should already be fairly well [documented](https://laowantong.github.io/paroxython/index.html):\n\n- [User manual](https://laowantong.github.io/paroxython/user_manual/index.html):\n  - [write a command pipeline to get recommendations](https://laowantong.github.io/paroxython/user_manual/index.html#pipeline-tutorial),\n  - [prepare your program collections for better results](https://laowantong.github.io/paroxython/user_manual/index#preparing-your-program-collection),\n  - [understand and modify the taxonomic classification](https://laowantong.github.io/paroxython/user_manual/index#taxonomy),\n  - and more.\n- [Developer manual](https://laowantong.github.io/paroxython/developer_manual/index.html):\n  - [get a rough idea of the program structure and operations](https://laowantong.github.io/paroxython/developer_manual/index.html#bird-view),\n  - [use the provided helpers to contribute to the code](helper-programs),\n  - and more.\n- [Module reference](https://laowantong.github.io/paroxython/#header-submodules).\n- [Feature specifications](https://github.com/laowantong/paroxython/blob/master/paroxython/resources/spec.md): a document mixing prose, tests, regular expressions and SQL queries to describe which algorithmic features are recognized and how.\n- [User types](https://github.com/laowantong/paroxython/blob/master/paroxython/user_types.py): all objects of interest are precisely typed and checked by [mypy](http://mypy-lang.org).\n\nFinally, a [battery of examples](https://github.com/laowantong/paroxython/tree/master/examples) and [comprehensive test coverage](https://github.com/laowantong/paroxython/tree/master/tests) should help answer any remaining question.\n',
    'author': 'Aristide Grange',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/laowantong/paroxython/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



r"""
Return a non-recursive, multiline dump of the AST (with some tweaks).

## Motivation and purpose

[Static analysis](https://en.wikipedia.org/wiki/Static_program_analysis) tools like Paroxython
generally start with the [AST](https://en.wikipedia.org/wiki/Abstract_syntax_tree) (abstract syntax
tree) of the program to analyze. For instance, when the string:

```python
'print("hello, world")'
```

... is fed to the function `parse()` of the modules [`ast`](https://docs.python.org/3/library/ast.html)
or [`typed_ast`](https://github.com/python/typed_ast), the following object is returned:

```plain
                           Module()
                              |
                              | body[0]
                              |
                            Expr()
                              |
                              | value
                              |
                            Call()
                             / \
                       func /   \ args[0]
                           /     \
Name(id='print', ctx=Load())     Str(s='hello, world', kind='')
```

The purpose of our function `flatten_ast()` is to transform this nested structure into a curated
pre-ordered sequence of text lines, each of which consisting in a key-value pair. A key is the
complete path from the root of the AST to either a node (e.g., `Name()`) or a node attribute
(e.g., `id`). The associated value comes after an equals sign (`=`):

```plain
/_type=Module
/body/_length=1
/body/1/_type=Expr
/body/1/_pos=1:1-
/body/1/value/_type=Call
/body/1/value/_hash=0x0001
/body/1/value/_pos=1:1-0-
/body/1/value/func/_type=Name
/body/1/value/func/_hash=0x0002
/body/1/value/func/_pos=1:1-0-0-
/body/1/value/func/id=print
/body/1/value/func/ctx/_type=Load
/body/1/value/args/_length=1
/body/1/value/args/1/_type=Str
/body/1/value/args/1/_hash=0x0003
/body/1/value/args/1/_pos=1:1-0-1-1-
/body/1/value/args/1/s=hello, world
/body/1/value/keywords/_length=0
/type_ignores/_length=0
```

Obviously, such a representation is quite verbose. But every single line encapsulates most of the
nesting information we need to understand its context. Since the recursion is linearly encoded,
requesting it does not require recursion any more. Thus,
[regular expressions](https://en.wikipedia.org/wiki/Regular_expression) become natural candidates
for the request language[^regex-recursion]. The current version of Paroxython explores this option.
The regular expression patterns are processed in `paroxython.parse_program` on the basis of the
definitions listed in `spec.md`.

[^regex-recursion]: Although nowadays, numerous flavours of regex engines support recursion,
the complexity of writing, reading and maintaining any pattern using it would quickly go through the
roof.

For instance, the following regular expression matches and localizes the nested function calls
(e.g., `foo(bar(42))`) without any recursion:

```re
           ^(.*)                    # Capture the path originating from the root...
                /_type=Call         #  ... to the nesting call.
\n(?:\1.+\n)*?\1                    # Skip any number of attributes of this call...
                /_pos=(?P<POS>.+)   # ... until we find and capture its line number.
\n(?:\1.+\n)* \1                    # Skip any number of...
                /args/.*            # ... arguments of this call...
                        /_type=Call # ... until we find a nested call.
```

When the first call is found, continuing in the subtree is a simple matter of restricting the search
to the block of lines that share the same prefix (referred by `\1`).

That's the general principle. In the next section, we enumerate the tweaks we make to the generation
of this text to make it more suitable for regular expression parsing.

## Implementation details

### Libraries used

When Paroxython is launched from Python 3.7 or lower, the parsing of source code into an AST is
performed by the third-party library [typed-ast](https://github.com/python/typed_ast). To quote the
README:

> `typed_ast` is a Python 3 package that provides a [...] parser similar to the standard `ast` library.
> Unlike `ast`, the parsers [...] are independent of the version of Python under which they are run.
> [...] `typed_ast` runs on CPython 3.5-3.8

So far so good. However:

> `typed_ast` will not be updated to support parsing Python 3.8 and newer.

This means that `typed_ast`, although convenient to parse Python 3.5 and newer, will chope on
Python 3.8 specific syntax, as assignment expressions.

This is the reason why, when launched from Python 3.8 or higher, Paroxython switches to the parser of
the standard library [`ast`](https://docs.python.org/3/library/ast.html). However, the ASTs generated
by the two libraries are not 100% identical. Wrapping their parsers requires us to erase as much
differences as possible. Some of the functions presented below are written for this purpose.

### On the fly: modification of the AST

#### Field reordering

Somewhat counterintuitively, in the original AST, several pieces of information pertaining to a
function signature or decoration are rejected _after_ the body of the function. As a result, we
lose the interesting property that the line numbers cannot decrease. We fix this by making `body`
the last field of a function or class definition
(see [issue #4](https://github.com/laowantong/paroxython/issues/4)).

#### Children numbering

A counter is inserted between each node and its children. Starting from 1 enables us to find the
last one by matching its number with the (previously captured) length of the list.

#### Field disambiguation

Accross the AST, the same fields are sometimes used in different contexts. To make it easier or
even possible to detect certain features, their usage is disambiguated by applying the following
renaming rules:

- `orelse`: used for both loop and conditional `else` branches. Replace the former with `loopelse`.
- `target`: used for assignment, comprehension, etc. Replace the former with `assigntarget`.
- `targets`: used for augmented assignment, deletion, etc. Replace the former with `assigntargets`.
- `value`: used for assignment (augmented or not), comprehension, assignment-expressions, etc.
  Replace the former ones with `assignvalue`.

Thus, in the output, there is no more need to search for the context of a field in the previous or
following lines.

#### Added fields

- `_type`: stores the name of the node (e.g., `Module`, `Expr`, `Call`).
- `_pos`: stores the line number (extracted from the attribute `lineno` of the node, if
  any[^lineno]), followed by a colon `:`, followed by the path from the root of the AST to this
  node. A path is coded as a sequence of positive integers, hyphen-separated and ended by a hyphen.
  Its fundamental property is: node \(n_2\) is nested within node \(n_1\) iff the path of \(n_1\)
  is a prefix of the path of \(n_2\).
- `_hash`: see next section.

[^lineno]: Numerous nodes, e.g. `Module`, `Store`, `Load`, `Lt`, `Add`, are not associated with a
    line number.

#### Expression decontextualization and hashing

In order to detect certain features, such as
[`swap`](https://repo/paroxython/resources/spec.md#feature-swap):

```python
(a[i], a[j]) = (a[j], a[i])
```
... we need to recognize a given expression accross its various occurrences. For this, we tweak the
output of the library function `ast.dump()`. For instance, in the following statement:

```python
a[i] = 42
```

... the expression `a[i]` is dumped as the string:

```python
"Subscript(value=Name(id='a', ctx=Load()), slice=Index(value=Name(id='i', ctx=Load())), ctx=Store())"
```

However, in:

```python
    return a[i]
```

... the same expression `a[i]` is dumped as the string:

```python
"Subscript(value=Name(id='a', ctx=Load()), slice=Index(value=Name(id='i', ctx=Load())), ctx=Load())"
```

As you can see, these two strings are not equal, due to the fact that the dump specifies not only
_which_ expressions are used, but _how_ they are used: this is the purpose of the field `ctx`. Once
this context is removed, both strings become:

```python
"Subscript(value=Name(id='a'), slice=Index(value=Name(id='i')))"
```

They can now be represented by the same hash value[^pseudo-hash], stored in the added field `_hash`.

[^pseudo-hash]:
    Actually, to avoid dealing with hash randomization in the unit tests, we use a function which
    merely increments a counter and associates it with any newly encountered string
    (see `PseudoHashFactory`).

### After the fact: modification of the flattened AST

#### Negative literal simplification

Normally, the AST structure of a numeric literal depends on its sign. For instance, 42 is parsed
as:
```python
Expr(value=Num(n=42))
```
... while -42 is parsed as:

```python
Expr(value=UnaryOp(op=USub(), operand=Num(n=42)))
```

To unify the further treatment of numbers, the function `simplify_negative_literals` post-processes
the generated flat representation to transform this kind of output:

```plain
/body/1/_type=Expr
/body/1/_pos=1:1-
/body/1/value/_type=UnaryOp
/body/1/value/_hash=0x0001
/body/1/value/_pos=1:1-0-
/body/1/value/op/_type=USub
/body/1/value/operand/_type=Num
/body/1/value/operand/_hash=0x0002
/body/1/value/operand/_pos=1:1-0-1-
/body/1/value/operand/n=42
```

into this one:

```plain
/body/1/_type=Expr
/body/1/_pos=1:1-
/body/1/value/_type=Num
/body/1/value/_hash=0x0001
/body/1/value/_pos=1:1-0-
/body/1/value/n=-42
```

#### Index simplification

Since Python 3.9, the class `ast.Index` is deprecated: simple indexes are represented by their value.

For instance, `a[42]` is parsed as:

```python
[...] slice=Constant(value=42) [...]
```

... instead of:

```python
[...] slice=Index(value=Constant(value=42))) [...]
```

To unify the further treatment of indexes, the function `simplify_indexes` post-processes the
generated flat representation to transform this kind of output:

```plain
/body/1/_type=Expr
/body/1/_pos=1:1-
/body/1/value/_type=Subscript
/body/1/value/_hash=0x0001
/body/1/value/_pos=1:1-0-
/body/1/value/value/_type=Name
/body/1/value/value/_hash=0x0002
/body/1/value/value/_pos=1:1-0-0-
/body/1/value/value/id=a
/body/1/value/value/ctx/_type=Load
/body/1/value/slice/_type=Index           # Line to be suppressed.
/body/1/value/slice/value/_type=Num       # Lines to be simplified...
/body/1/value/slice/value/_hash=0x0003    # ...
/body/1/value/slice/value/_pos=1:1-0-1-0- # ...
/body/1/value/slice/value/n=42            # ...
/body/1/value/ctx/_type=Load
```

into this one:

```plain
/body/1/_type=Expr
/body/1/_pos=1:1-
/body/1/value/_type=Subscript
/body/1/value/_hash=0x0001
/body/1/value/_pos=1:1-0-
/body/1/value/value/_type=Name
/body/1/value/value/_hash=0x0002
/body/1/value/value/_pos=1:1-0-0-
/body/1/value/value/id=a
/body/1/value/value/ctx/_type=Load
/body/1/value/slice/_type=Num        # Simplified lines...
/body/1/value/slice/_hash=0x0003     # ...
/body/1/value/slice/_pos=1:1-0-1-    # ... (the last segment of the path is dropped)
/body/1/value/slice/n=42             # ...
/body/1/value/ctx/_type=Load
```

#### Suppressing useless lines

- Starting with Python 3.8, `kind` is an attribute of a constant allowing tools to distinguish
certain flavors of the strings . Its possible values are `"rf"`, `"f"`, `"u"` and `""` in the
third-party library `typed_ast`, but `"u"` and `None` in the Python 3.8 standard library `ast`.
Since such details are of no interest to us, we don't bother making them consistent. The function
`suppress_kinds` deletes from the flat AST any line ending with `kind=.*`.
- Python 3.8 introduces [positional-only parameters](https://www.python.org/dev/peps/pep-0570/)
to respond to “certain challenges for library authors and users of APIs”. This is far beyond our
pedagogical scope: we are afraid we have no choice but to sell these lines for scientific
experiments (`suppress_posonlyargs`).

#### Reverting the representation of constants to the good ol' days

Probably for excellent reasons, Python 3.8 has decided that `Num(n)`, `Str(s)`, `Bytes(s)`,
`Ellipsis` and `NameConstant(value)` should all be unified under `Constant(value)`. As far as we
are concerned, this change makes the type of the constants much harder to determine. Until we
know more, we are restoring the previous representations with `backport_all_constants`.

#### Unquoting strings

`unquote()` suppress the single or double quotes delimiters after `=`. This makes for simpler regular
expressions in `spec.md`.

#### Correcting the span of decorated callable

`correct_all_decorated_callable_starts()` fixes a bug of `typed_ast` in which decorated functions
and classes would incorrectly be marked as starting from the line of their first decorator.
"""


import sys
from typing import Any, Callable, Match

import regex  # type: ignore


def select_ast_post_processing():
    """Customize the post-treatment of the AST depending on the parser used to create it."""

    if sys.version_info >= (3, 9):

        def post_process(flat_ast):
            flat_ast = suppress_kinds(flat_ast)
            flat_ast = suppress_posonlyargs(flat_ast)
            flat_ast = backport_all_constants(flat_ast)
            flat_ast = simplify_negative_literals(flat_ast)
            flat_ast = unquote(flat_ast)
            return flat_ast

    elif sys.version_info >= (3, 8):

        def post_process(flat_ast):
            flat_ast = suppress_kinds(flat_ast)
            flat_ast = suppress_posonlyargs(flat_ast)
            flat_ast = backport_all_constants(flat_ast)
            flat_ast = simplify_negative_literals(flat_ast)
            flat_ast = simplify_indexes(flat_ast)
            flat_ast = unquote(flat_ast)
            return flat_ast

    else:

        def post_process(flat_ast):
            flat_ast = suppress_kinds(flat_ast)
            flat_ast = simplify_negative_literals(flat_ast)
            flat_ast = simplify_indexes(flat_ast)
            flat_ast = unquote(flat_ast)
            flat_ast = correct_all_decorated_callable_starts(flat_ast)
            return flat_ast

    return post_process


# fmt: off
if sys.version_info >= (3, 8):
    import ast
else:
    from typed_ast import ast3 as ast # type: ignore
post_process = select_ast_post_processing()
# fmt: on


def flatten_ast(tree) -> str:
    """Return a non-recursive, multiline dump of the AST (with some tweaks).

    A simple wrapper around the recursive function `flatten_node`, with post-processing treatments
    depending on the AST library (either the standard `ast` module or the third-party `typed_ast`).
    """
    pseudo_hash.reset()
    flat_ast = flatten_node(tree)
    flat_ast = post_process(flat_ast)
    return flat_ast


def flatten_node(
    node: Any,
    prefix: str = "",
    path: str = "",
    remove_context: Callable = regex.compile(r", ctx=.+?\(\)").sub,
) -> str:
    r"""Traverse recursively (in pre-order) the given AST node and flatten its subtree.

    Args:
        node (Any): The node to traverse. Initially, the whole AST.
        prefix (str, optional): The prefix of the current line to dump. Defaults to `""`.
        path (str, optional): The path of the current node. Defaults to `""`.
        remove_context (Callable, optional): A function removing the node context encoded in the
            result of `ast3.dump()`.
            [Not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
            Defaults to `regex.compile(r", ctx=.+?\(\)").sub`.

    Returns:
        str: A flat representation of the given node.
    """
    if isinstance(node, ast.AST):
        acc = [f"{prefix}/_type={type(node).__name__}\n"]
        if isinstance(node, ast.expr):
            node_repr = remove_context("", ast.dump(node))
            acc.append(f"{prefix}/_hash={pseudo_hash(node_repr)}\n")
        if "lineno" in node._attributes:
            acc.append(f"{prefix}/_pos={node.lineno}:{path[2:]}\n")
        fields = ast.iter_fields(node)
        if isinstance(node, (ast.FunctionDef, ast.ClassDef)):
            # make `body` the last item of a `def` clause
            fields = iter(sorted(fields, key=lambda c: c[0] == "body"))
        for (i, (name, x)) in enumerate(fields):
            if name == "orelse" and isinstance(node, (ast.For, ast.While, ast.AsyncFor)):
                name = "loopelse"  # make the `orelse` clause specific to conditionals
            elif name == "targets" and isinstance(node, ast.Assign):
                name = "assigntargets"  # `targets` is also used for `Delete`, etc.
            elif name == "target" and isinstance(node, ast.AugAssign):
                name = "assigntarget"  # `target` is also used for comprehension, etc.
            elif name == "value" and isinstance(node, (ast.Assign, ast.AugAssign)):
                name = "assignvalue"  # `value` is also used for comprehension, etc.
            acc.append(flatten_node(x, f"{prefix}/{name}", f"{path}{i}-"))
        return "".join(acc)
    elif isinstance(node, list):
        acc = [f"{prefix}/_length={len(node)}\n"]
        for (i, x) in enumerate(node, 1):  # number the children from 1
            acc.append(flatten_node(x, f"{prefix}/{i}", f"{path}{i}-"))
        return "".join(acc)
    else:
        # If the node is a terminal value, dump it
        return f"""{prefix}={repr(node)}\n"""


def simplify_negative_literals(
    flat_ast: str,
    sub: Callable = regex.compile(
        r"""(?mx)
                      ^(.*?)/_type=UnaryOp
            (\n(?:.+\n)*?)\1/op/_type=USub
             \n(?:.+\n)*? \1/operand/n=(.+)
        """
    ).sub,
) -> str:
    """[Simplify the negative literals](#negative-literal-simplification) of a given flat AST.

    Argument `sub` [not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
    """
    return sub(r"\1/_type=Num\2\1/n=-\3", flat_ast)


def simplify_indexes(
    flat_ast: str,
    subn_index: Callable = regex.compile(r"(?m)^(.*?)/_type=Index\n").subn,
    sub_slice_value_pos: Callable = regex.compile(r"(?m)^([^=]*/slice/value/_pos=.+)\d+-").sub,
    sub_slice_value: Callable = regex.compile(r"(?m)^([^=]*/slice)/value").sub,
) -> str:
    """[Simplify the indexes](#index-simplification) of a given flat AST.

    Argument `sub` [not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
    """
    (flat_ast, n) = subn_index("", flat_ast)
    if n == 0:
        return flat_ast
    flat_ast = sub_slice_value_pos(r"\1", flat_ast)
    return sub_slice_value(r"\1", flat_ast)


def backport_all_constants(
    flat_ast: str,
    sub: Callable = regex.compile(
        r"""(?mx)
            ^(.*?)/_type=Constant
            ((?:\n.+)+)
              \n\1/value=(.+)
        """
    ).sub,
) -> str:
    """Restore the more detailed representation of constants used before Python 3.8.

    Argument `sub` [not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
    """
    return sub(replace_one_constant, flat_ast)


def replace_one_constant(m: Match) -> str:
    """Define the replacement function used in `backport_all_constants`."""
    if m[3].startswith(("'", '"')):
        return f"{m[1]}/_type=Str{m[2]}\n{m[1]}/s={m[3]}"
    elif m[3] in ("True", "False", "None"):
        return f"{m[1]}/_type=NameConstant{m[2]}\n{m[1]}/value={m[3]}"
    if m[3].startswith("b'"):
        return f"{m[1]}/_type=Bytes{m[2]}\n{m[1]}/s={m[3]}"
    if m[3] == "Ellipsis":
        return f"{m[1]}/_type=Ellipsis{m[2]}"
    else:
        return f"{m[1]}/_type=Num{m[2]}\n{m[1]}/n={m[3]}"


def correct_all_decorated_callable_starts(
    flat_ast: str,
    sub: Callable = regex.compile(
        r"""(?mx)
                     ^(.*?)/_type=(FunctionDef|ClassDef)
            (\n(?:.+\n)*?\1/_pos=)\d+:(.*
            \n(?:.+\n)*?\1/decorator_list/_length=([^0]\d*)
            \n(?:.+\n)*?\1/decorator_list/1/_pos=(\d+))
        """
    ).sub,
) -> str:
    """Make the span of a decorated callable starting from `def`, not from the first decorator.

    ..note::
        The bug is present in the third-party module `typed_ast`, but fixed in Python 3.8's standard
        library `ast`.

    Argument `sub` [not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
    """
    return sub(replace_one_decorated_callable_start, flat_ast)


def replace_one_decorated_callable_start(m: Match) -> str:
    """Define the replacement function used in `correct_all_decorated_callable_starts`."""
    start = int(m[5]) + int(m[6])  # number of decorators + line number of the first decorator
    return f"{m[1]}/_type={m[2]}{m[3]}{start}:{m[4]}"


def unquote(
    flat_ast: str,
    sub: Callable = regex.compile(r"""=["'](.*)['"]\n""").sub,
) -> str:
    """Suppress the quote delimiters after “=“ to simplify the regular expressions of `spec.md`.

    Argument `sub` [not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
    """
    return sub(r"=\1\n", flat_ast)


def suppress_kinds(
    flat_ast: str,
    sub: Callable = regex.compile(r"(?m)^.+/kind=.*\n").sub,
) -> str:
    """Suppress all leaves containing `/kind=`. They are useless in the following.

    Argument `sub` [not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
    """
    return sub("", flat_ast)


def suppress_posonlyargs(
    flat_ast: str,
    sub: Callable = regex.compile(r"(?m)^.+/args/posonlyargs/_length=\d+\n").sub,
) -> str:
    """Starting from Python 3.8, suppress positional-only parameters in function definitions.

    Argument `sub` [not to be explicitly provided.](developer_manual/index.html#default-argument-trick)
    """
    return sub("", flat_ast)


__pdoc__ = {"PseudoHashFactory.__call__": True}


class PseudoHashFactory:
    """A factory that creates a deterministic replacement for the `hash()` builtin function."""

    def __init__(self):
        self.reset()

    def reset(self):
        """Restart the global counter and empties the cache."""
        self.i = 0
        self.cache = {}

    def __call__(self, x):
        """Return a 4-digits hexadecimal identifier for the given immutable value."""
        if x not in self.cache:
            self.i += 1
            self.cache[x] = f"0x{self.i:04x}"
            # 4.294.967.296 distinct expressions per program ought to be enough for anybody.
        return self.cache[x]


pseudo_hash = PseudoHashFactory()

if __name__ == "__main__":
    Path = __import__("pathlib").Path
    source = Path("sandbox/source.py").read_text()
    print(flatten_ast(ast.parse(source)))



r"""Assess the learning cost associated with the introduction of the given programs.

Each program has been previously found to implement a number of notions represented by a list of
taxa of the form: \(\textrm{edge}_0/\textrm{edge}_1/.../\textrm{edge}_n\), e.g.,
`flow/loop/exit/early/break`.

Furthermore, the notions already imparted may cover a certain **prefix** of such a taxon, here, e.g.
`flow/loop`. The learning cost should therefore not take this into account.

The cost of the remaining edges, here `exit/early/break`, is approximated by a function taking
two zero-based indexes `start` (inclusive) and `stop` (exclusive), and returning a floating number. Currently, only two such functions are provided:

1. `range_to_cost_linear()`, which simply counts the number of edges;
2. `range_to_cost_zeno()`, which calculates the sum of a decreasing function of the positions
   (with the assumption that the introduction of a sub-concept costs half that of its parent concept).
"""

from functools import lru_cache

from .user_types import (
    AssessedPrograms,
    AssessmentStrategy,
    ProgramInfos,
    ProgramPathSet,
    TaxonName,
    TaxonNameSet,
)

__pdoc__ = {
    "LearningCostAssessor.__call__": True,
    "LearningCostAssessor.__init__": True,
    "LearningCostAssessor": "",
}


@lru_cache(maxsize=None)  # NB: memoization needed for consistency with mypy's typing
def range_to_cost_linear(start: int, stop: int) -> float:
    """Return the length of the slice between `start` (inclusive) and `stop` (exclusive)."""
    return float(stop - start)


@lru_cache(maxsize=None)
def range_to_cost_zeno(start: int, stop: int) -> float:
    r"""Calculate the sum of the given slice of a [Zeno's geometric series](https://en.wikipedia.org/wiki/Zeno%27s_paradoxes#Dichotomy_paradox).

    Specifically: $$\sum_{i=\textrm{start}}^{\textrm{stop}-1}{\frac{1}{2^{i+1}}}$$

    Since it converges absolutely towards 1, the returned value is in the interval \([0, 1[\).

    <center><a title="Jim.belk / Public domain" href="https://commons.wikimedia.org/wiki/File:Geometric_Segment.svg"><img width="256" alt="Geometric Segment" src="https://upload.wikimedia.org/wikipedia/commons/thumb/a/ab/Geometric_Segment.svg/256px-Geometric_Segment.svg.png"></a></center>

    Example:
        If `flow/loop` is already imparted, the learning cost of `flow/loop/exit/early/break` will
        be:

    >>> range_to_cost_zeno(2, 5)
    0.21875

    Indeed,
    \(\frac{1}{2^3}+\frac{1}{2^4}+\frac{1}{2^5}=\frac{1}{8}+\frac{1}{16}+\frac{1}{32}=0.21875\).

    """
    return sum(2 ** ~i for i in range(start, stop))


class LearningCostAssessor:
    """Evaluate the learning costs of programs with respect to the given imparted knowledge."""

    def __init__(self, programs: ProgramInfos, assessment_strategy: AssessmentStrategy = "zeno"):
        """Set the programs and the function to be used for the learning cost calculation.

        Args:
            assessment_strategy (str, optional): Either:

                - `"zeno"`: `range_to_cost_zeno` (default).
                - `"linear"`: `range_to_cost_linear`.

        Raises:
            NotImplementedError: Raised in case of unknown `assessment_strategy`.
        """
        self.programs = programs
        if assessment_strategy.lower() == "zeno":
            self.range_to_cost = range_to_cost_zeno
        elif assessment_strategy == "linear":
            self.range_to_cost = range_to_cost_linear
        else:
            raise NotImplementedError
        self.range_to_cost.cache_clear()
        self.taxon_cost.cache_clear()

    def set_imparted_knowledge(self, imparted_knowledge: TaxonNameSet) -> None:
        self.imparted_knowledge = imparted_knowledge

    @lru_cache(maxsize=None)
    def taxon_cost(self, taxon: TaxonName) -> float:
        """Evaluate the learning cost of a given taxon.

        The bounds of the not-yet-imparted suffix of the taxon are extracted and passed to the cost
        assessment function.

        .. note::
          The learning cost of a taxon prefixed by `"meta/"` is assumed to be zero.
        """
        if taxon.startswith("meta/"):
            return 0
        (start, stop) = (0, 0)
        if taxon not in self.imparted_knowledge:
            edges = taxon.split("/")
            stop = len(edges)
            for start in range(stop - 1, -1, -1):
                if "/".join(edges[:start]) in self.imparted_knowledge:
                    break
        return self.range_to_cost(start, stop)

    def __call__(self, selected_programs: ProgramPathSet) -> AssessedPrograms:
        """Associate the given selected programs with the total learning cost of their taxa.

        Args:
            selected_programs (ProgramPathSet): A set of program paths.

        Returns:
            AssessedPrograms: A list of tuples `(total_cost, ProgramPath)` sorted by increasing cost.
        """
        result = []
        for program_path in selected_programs:
            total_cost = 0.0
            for taxon_name in self.programs[program_path]["taxa"]:
                total_cost += self.taxon_cost(taxon_name)
            result.append((total_cost, program_path))
        return sorted(result)
